from argparse import ArgumentParser, RawDescriptionHelpFormatter
import base64
import json
from pathlib import Path
import pickle
import os

import pandas as pd
import plotly.express as px

from as_analytics.lsreport import read_filetree_ls


def _get_parser():
    parser = ArgumentParser(
        description="Parser for output of `tools/lstraverse.sh`.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "path",
        type=str,
        help="Path of filtree as output from `tools/lstraverse.sh`.",
    )
    parser.add_argument(
        "--expid",
        type=str,
        help="Expid. Only for saving purposes",
    )
    parser.add_argument(
        "--root",
        type=str,
        default="",
        help="Root directory to examine.",
    )
    parser.add_argument(
        "--figure",
        type=str,
        default=None,
        choices=["treemap", "sunburst", "icicle", "all", None],
        help="Root directory to examine.",
    )
    parser.add_argument(
        "--show",
        action="store_true",
        help="Whether to show the figure(s) after plotting.",
    )
    parser.add_argument(
        "--keep-dots",
        action="store_true",
        help="Whether to keep the entries for '.' and '..' in the original dataframe.",
    )
    parser.add_argument("--cache", action="store_true", help="Read / write to ~/.cache")
    return parser


def _json_serialize(obj):
    if isinstance(obj, Path):
        return str(obj)
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


def _get_cache_path(*args, **kwargs) -> Path:
    key = json.dumps([args, kwargs], default=_json_serialize)
    encoded = base64.b64encode(key.encode("utf-8"))
    file = encoded.decode("utf-8")
    directory = os.path.basename(__file__)
    return Path.home().resolve() / ".cache" / directory / file


def _cache_read(cache, /, *args, **kwargs):
    if cache is False:
        return
    cpath = _get_cache_path(*args, **kwargs)
    if not cpath.exists():
        return None
    print("Reading from cache...")
    with open(cpath, "rb") as fb:
        return pickle.load(fb)


def _cache_write(cache, obj, /, *args, **kwargs):
    if cache is False:
        return
    print("Writing to cache...")
    cpath = _get_cache_path(*args, **kwargs)
    if not cpath.parent.exists():
        cpath.parent.mkdir(parents=True)
    with open(cpath, "wb") as fb:
        pickle.dump(obj, fb)


def _get_subtree_df(df, root):
    # Normalize Parent paths
    df = df.copy()
    df["Parent"] = df["Parent"].str.rstrip("/")

    # Compute common root
    common_root = os.path.commonpath(list(df["Parent"]))
    print(f"\nCommon root: {common_root}")

    # Normalize requested root
    if root in (None, ""):
        abs_root = common_root
    else:
        # If root is absolute, use it directly
        if os.path.isabs(root):
            abs_root = os.path.normpath(root)
        else:
            # If root is relative, join it with common root
            abs_root = os.path.normpath(os.path.join(common_root, root))
    print(f"Absolute root: {abs_root}")

    # Filter to subtree - only include files that are under the absolute root
    mask = df["Parent"].str.startswith(abs_root)
    subtree_df = df[mask].copy()

    if subtree_df.empty:
        raise ValueError(
            f"No entries found under root '{root}' (resolved to '{abs_root}')."
        )

    # Simply use the Parent path and Filename
    subtree_df["full_path"] = subtree_df.apply(
        lambda row: [str(part) for part in row["Parent"].split("/") if part]
        + [row["Filename"]],
        axis=1,
    )
    subtree_df["name"] = subtree_df["Filename"]

    return subtree_df


def draw(df: pd.DataFrame, root: str = None, figure: str = "treemap"):
    """
    Visualize file structure under a given root directory using Plotly.

    Parameters:
        df (pd.DataFrame): The DataFrame containing ls -la data. Must have 'Parent' and 'Filename'.
        root (str | None): Subdirectory to start visualization from. Can be absolute or relative to the common root.
        figure (str): One of 'treemap', 'sunburst', or 'icicle'.
    """
    if figure not in {"treemap", "sunburst", "icicle"}:
        raise ValueError(
            f"Unsupported figure '{figure}'. Use 'treemap', 'sunburst', or 'icicle'."
        )

    if "Parent" not in df.columns:
        raise ValueError("DataFrame must include a 'Parent' column.")

    if "Filename" not in df.columns:
        raise ValueError("DataFrame must include a 'Filename' column.")

    # Compute directory sizes
    sizes, parents = compute_directory_size(df)

    # Create visualization DataFrame
    viz_df = pd.DataFrame(
        {
            "name": list(sizes.keys()),
            "parent": list(parents.values()),
            "value": list(sizes.values()),
        }
    )

    # Filter by root if specified
    if root:
        mask = viz_df["name"].str.startswith(root)
        viz_df = viz_df[mask]
        if viz_df.empty:
            raise ValueError(f"No entries found under root '{root}'.")

    # Create the figure
    fig = getattr(px, figure)(
        viz_df,
        names="name",
        parents="parent",
        values="value",
        hover_data=["name", "value"],
    )

    # Get the root for the title
    abs_root = root if root else viz_df[viz_df["parent"] == ""]["name"].iloc[0]

    fig.update_traces(root_color="lightgrey")
    fig.update_layout(
        title=f"{figure.title()} of directory sizes under: {abs_root}",
        margin=dict(t=40, l=0, r=0, b=0),
    )
    return fig


def makefigure(df, root, figure, show, save, prefix, cache, **kwargs):

    if figure == "all":
        return [
            makefigure(df, root, "treemap", show, save, prefix, cache, **kwargs),
            makefigure(df, root, "sunburst", show, save, prefix, cache, **kwargs),
            makefigure(df, root, "icicle", show, save, prefix, cache, **kwargs),
        ]

    print(f"Plotting {figure}.")
    fig = _cache_read(cache, "figure", root=root, figure=figure, **kwargs)
    if fig is None:
        fig = draw(df, root, figure)
        _cache_write(cache, "figure", root=root, figure=figure, **kwargs)
    if save:
        fig.write_image(f"figures/{prefix}{figure}.png")
    if show:
        fig.show()
    return fig


def compute_directory_size(df):
    df["Path"] = (df["Parent"] + "/" + df["Filename"]).apply(
        lambda x: x.removeprefix("/")
    )
    sizes = {}
    parents = {}
    for directory, directory_df in df.groupby("Parent"):
        sizes[directory] = int(directory_df["Filesize"].sum())
        if "/" not in directory:
            parents[directory] = ""
        else:
            parent_dir, _ = directory.rsplit("/", 1)
            parents[directory] = parent_dir
    return sizes, parents


def main():
    args = _get_parser().parse_args()
    path = Path(args.path).resolve()
    expid = args.expid
    root = args.root
    figure = args.figure
    show = args.show
    keep_dots = args.keep_dots
    cache = args.cache

    df = None

    df = _cache_read(cache, "df", path=path, expid=expid, keep_dots=keep_dots)
    if df is None:
        df = read_filetree_ls(path, join=True, common=True, keep_dots=keep_dots)
        _cache_write(cache, df, "df", path=path, expid=expid, keep_dots=keep_dots)

    df["Path"] = df["Path"] = df["Parent"] + "/" + df["Filename"]

    if figure is not None:
        makefigure(
            df=df,
            root=root,
            figure=figure,
            show=show,
            save=True,
            prefix=(f"{expid}_" if expid else ""),
            cache=cache,
            path=path,
            expid=expid,
            keep_dots=keep_dots,
        )


if __name__ == "__main__":
    main()
